#include <stdint.h>

// hardcoded nand chip configs that can be used if the board is controlled by
// a bare terminal emulator instead of the desktop application
/*
// K9F2G08U0M 0xECDA8015     mb=256  page=2048 oob=64  block=128KB  bus=x8  bad="1st or 2nd page (1st byte in OOB = non FFh)"  sp1=TRUE sp2=TRUE  slp=FALSE slpm2=FALSE sb1=TRUE sb6=FALSE ecc="? (1-bit correct, 2-bit detect)" bpc=1 idlen=4
uint32_t g_unSizeID = 4;
uint32_t g_unSizePage = 2048;
uint32_t g_unSizePageOOB = 64;
uint32_t g_unSizeBlock = 128 * 1024;
uint32_t g_unNumBlocks = 2048;
*/
/*
// K9GAG08U0E 0xECD584725042 mb=2048 page=8192 oob=436 block=1024KB bus=x8  bad="1st or last page (1st byte in OOB = non FFh)" sp1=TRUE sp2=FALSE splp=TRUE slpm2=FALSE sb1=TRUE sb6=FALSE ecc="24-bit/1078.5-byte" bpc=2 idlen=6
uint32_t g_unSizeID = 6;
uint32_t g_unSizePage = 8192;
uint32_t g_unSizePageOOB = 436;
uint32_t g_unSizeBlock = 1024 * 1024;
uint32_t g_unNumBlocks = 2076;
*/

uint32_t g_unSizeID = 0;
uint32_t g_unSizePage = 0;
uint32_t g_unSizePageOOB = 0;
uint32_t g_unSizeBlock = 0;
uint32_t g_unNumBlocks = 0;

// these are calculated once after the nand chip config has been set
uint32_t g_unSizePageRaw = 0;
uint32_t g_unPagesPerBlock = 0;
uint32_t g_unColAddrBytes = 0;
uint32_t g_unRowAddrBytes = 0;
uint32_t g_unAddrBytes = 0;
uint32_t g_unColAddrBits = 0;
uint32_t g_unRowPageAddrBits = 0;
uint32_t g_unRowBlockAddrBits = 0;

// bluepill to nand chip wiring
#define RB   8 // PA8
#define RE   9 // PA9
#define CE  10 // PA10
#define CLE 15 // PA15
#define ALE 19 // PB3
#define WE  20 // PB4
#define WP  22 // PB6
#define IO0 24 // PB8
#define IO1 25 // PB9
#define IO2 26 // PB10
#define IO3 27 // PB11
#define IO4 28 // PB12
#define IO5 29 // PB13
#define IO6 30 // PB14
#define IO7 31 // PB15

// nand chip access functions
void nandSetup();
void nandReset();
void nandEnable();
void nandDisable();
uint8_t nandReadDataByte();
void nandWriteDataByte(uint8_t data);
void nandWriteCommandByte(uint8_t cmd);
void nandWriteAddressByte(uint8_t addr);
void nandWriteAddress(uint32_t unRow, uint32_t unColumn);
void nandWriteAddress(uint32_t unBlock, uint32_t unPage, uint32_t unColumn);
void nandSetDataDirOut();
void nandSetDataDirIn();

uint32_t countStorageBits(uint32_t n) 
{ 
  if(n == 0)
  {
    return 0;
  }
  
  n--;
  uint32_t count = 0;
  while(n) 
  { 
    count++; 
    n >>= 1; 
  } 
  return count; 
}

void printHex8(uint8_t data)
{
  if(data < 0x10)
  {
    Serial.write(0x30);
  }
  Serial.print(data, HEX);
}

void busyLedOn()
{
  digitalWrite(LED_BUILTIN, 0);
}

void busyLedOff()
{
  digitalWrite(LED_BUILTIN, 1);
}

void applyNandConfig()
{
  // if any config has been set yet...
  if(g_unSizeID)
  {
    g_unSizePageRaw = g_unSizePage + g_unSizePageOOB;
    g_unPagesPerBlock = g_unSizeBlock / g_unSizePage;
    g_unColAddrBits = countStorageBits(g_unSizePageRaw);
    g_unRowBlockAddrBits = countStorageBits(g_unNumBlocks);
    g_unRowPageAddrBits = countStorageBits(g_unPagesPerBlock);
    g_unColAddrBytes = (g_unColAddrBits / 8) + 1;
    g_unRowAddrBytes = ((g_unRowBlockAddrBits + g_unRowPageAddrBits) / 8) + 1;
    g_unAddrBytes = g_unColAddrBytes + g_unRowAddrBytes;
  }  
}

void setup()
{
  // disable JTAG so PB3/PB4/PA15 are free to use
  afio_cfg_debug_ports(AFIO_DEBUG_SW_ONLY);
  // enable USB CDC (baud rate doesn't matter)
  Serial.begin(115200);
  pinMode(LED_BUILTIN, OUTPUT);
  busyLedOff();

  // apply any chip config that has been hardcoded at the top of this file
  applyNandConfig();

  // set up the chip pins and initialize it
  nandSetup();
}

uint32_t readArgs(const uint32_t*& pArgs)
{
  static uint32_t rgArgs[21];
  uint32_t unArgCnt = 0;
  char szArgBuf[11];
  uint32_t unArgLen = 0;
  char c = ' ';
  do
  {
    if(Serial.available())
    {
      c = Serial.read();
      if(c >= '0' && c <= '9')
      {
        if(unArgLen == sizeof(szArgBuf) - 1)
        {
          return 0;
        }          
        szArgBuf[unArgLen++] = c;
      }
      else if(c == ',' || c == '\r' || c == '\n')
      {
          if(unArgLen > 0)
          {
            szArgBuf[unArgLen] = 0;
            rgArgs[unArgCnt++] = (uint32_t)atoi(szArgBuf);
            unArgLen = 0;
          }
      }
    }
  } 
  while((c != '\r') && (c != '\n') && (unArgCnt < 21));

  pArgs = rgArgs;
  return unArgCnt;
}

bool checkChipConfigSet()
{
  if(g_unSizeID == 0)
  {
    Serial.println("ERR,chip config not set");
    return false;       
  }
  return true;
}

bool checkMinArgs(uint32_t unArgNum, uint32_t unMinArgs)
{
  if(unArgNum < unMinArgs)
  {
    Serial.println("ERR,not enough arguments");
    return false;       
  }
  return true;
}

void transferNandPageToSerial(uint32_t unBlock, uint32_t unPage, uint32_t unStartColumn, bool bBinary)
{
  busyLedOn();
 
  nandSetDataDirOut();
  nandWriteCommandByte(0x00);
  nandWriteAddress(unBlock, unPage, unStartColumn);
  nandWriteCommandByte(0x30);
  nandBusyWait();
  nandSetDataDirIn();
  
  for(uint32_t i = unStartColumn; i < (g_unSizePageRaw); i++)
  {
    uint8_t db = nandReadDataByte();
    if(bBinary)
    {
      Serial.write(db);
    }
    else
    {
      printHex8(db);          
    }
  }
  if(!bBinary)
  {
    Serial.println("");
  }

  busyLedOff();
}

void execCmdGetChipConfig(const uint32_t* pArgs, uint32_t unArgNum)
{  
  Serial.print("OK,"); Serial.print(g_unSizeID, DEC);
  Serial.print(","); Serial.print(g_unSizePage, DEC);
  Serial.print(","); Serial.print(g_unSizePageOOB, DEC);
  Serial.print(","); Serial.print(g_unSizeBlock, DEC);
  Serial.print(","); Serial.print(g_unNumBlocks, DEC);
  Serial.println("");
}

void execCmdSetChipConfig(const uint32_t* pArgs, uint32_t unArgNum)
{
  if(!checkMinArgs(unArgNum, 5))
  {
    return;
  }
  
  g_unSizeID = pArgs[0];
  g_unSizePage = pArgs[1];
  g_unSizePageOOB = pArgs[2];
  g_unSizeBlock = pArgs[3];
  g_unNumBlocks = pArgs[4];
  applyNandConfig();
  Serial.println("OK");
}

void execCmdGetDebugInfo(const uint32_t* pArgs, uint32_t unArgNum)
{
  Serial.print("OK,"); Serial.print(g_unRowBlockAddrBits, DEC);
  Serial.print(","); Serial.print(g_unRowPageAddrBits, DEC);
  Serial.print(","); Serial.print(g_unRowAddrBytes, DEC);
  Serial.print(","); Serial.print(g_unColAddrBits, DEC);
  Serial.print(","); Serial.print(g_unColAddrBytes, DEC);
  Serial.println("");
}

void execCmdReadID(const uint32_t* pArgs, uint32_t unArgNum)
{
  if(!checkChipConfigSet())
  {
    return;
  }

  nandEnable();
  nandWriteCommandByte(0x90);
  nandWriteAddressByte(0x00);
  nandSetDataDirIn();

  Serial.print("OK,0x");
  for(uint32_t i = 0; i < g_unSizeID; i++)
  {
    printHex8(nandReadDataByte());
  }
  Serial.println("");
  nandDisable();
}

void execCmdNandReadPage(const uint32_t* pArgs, uint32_t unArgNum)
{
  if(!checkChipConfigSet())
  {
    return;
  }

  if(!checkMinArgs(unArgNum, 4))
  {
    return;
  }

  nandEnable();
  transferNandPageToSerial(pArgs[0], pArgs[1], pArgs[2], pArgs[3] != 0);
  nandDisable();
  Serial.println("OK");    
}

void execCmdNandReadBlock(const uint32_t* pArgs, uint32_t unArgNum)
{
  if(!checkChipConfigSet())
  {
    return;
  }

  if(!checkMinArgs(unArgNum, 1))
  {
    return;
  }

  nandEnable();
  for(uint32_t i = 0; i < g_unPagesPerBlock; i++)
  {
    transferNandPageToSerial(pArgs[0], i, 0, true);
  }
  nandDisable();
  Serial.println("OK");    
}

void execCmdNandReset(const uint32_t* pArgs, uint32_t unArgNum)
{
  if(!checkChipConfigSet())
  {
    return;
  }

  nandReset();
  Serial.println("OK");    
}

void loop()
{
  const uint32_t* pArgs;
  uint32_t unArgNum = readArgs(pArgs);
  if(unArgNum == 0)
  {
    return;
  }

  switch(pArgs[0])
  {
    case 1:
    {
      execCmdGetChipConfig(++pArgs, --unArgNum);
      break;
    }
    case 2:
    {
      execCmdSetChipConfig(++pArgs, --unArgNum);
      break;
    }
    case 3:
    {
      execCmdReadID(++pArgs, --unArgNum);
      break;
    }
    case 4:
    {
      execCmdNandReadPage(++pArgs, --unArgNum);
      break;
    }
    case 5:
    {
      execCmdNandReadBlock(++pArgs, --unArgNum);
      break;
    }
    case 6:
    {
      execCmdNandReset(++pArgs, --unArgNum);
      break;
    }
    case 666:
    {
      execCmdGetDebugInfo(++pArgs, --unArgNum);
      break;
    }
    default:
    {
      Serial.println("ERR,unknown command");
      break;
    }
  }
}

void nandSetup()
{
  pinMode(WP, OUTPUT);
  digitalWrite(WP, 0);
  pinMode(CE, OUTPUT);
  digitalWrite(CE, 1);
  pinMode(CLE, OUTPUT);
  pinMode(ALE, OUTPUT);
  pinMode(RE, OUTPUT);
  pinMode(WE, OUTPUT);
  pinMode(RB, INPUT_PULLUP); // RB is open drain
  nandReset();  
}

void nandReset()
{
  // send reset command
  nandSetDataDirOut();
  digitalWrite(CE, 0);
  nandWriteCommandByte(0xff);
  // wait at least 500us
  delayMicroseconds(1000);
  // set chips initial state
  digitalWrite(CE, 1);
  digitalWrite(CLE, 0);
  digitalWrite(ALE, 0);
  digitalWrite(WE, 1);
  digitalWrite(RE, 1);
}

void nandEnable()
{
  nandSetDataDirOut();
  digitalWrite(CLE, 0);
  digitalWrite(ALE, 0);
  digitalWrite(RE, 1);
  digitalWrite(WE, 1);
  digitalWrite(CE, 0);
}

void nandDisable()
{
  digitalWrite(CE, 1);
  digitalWrite(CLE, 0);
  digitalWrite(ALE, 0);
  digitalWrite(RE, 1);
  digitalWrite(WE, 1);
}

void nandBusyWait()
{
  // WE high to busy time (tWB), max. 100ns
  asm("NOP");
  asm("NOP");
  asm("NOP");
  asm("NOP");
  asm("NOP");
  asm("NOP");
  asm("NOP");
  // wait for RB to go high
  while((GPIOA->regs->IDR & 0x100) == 0)
  {    
  }
}

uint8_t nandReadDataByte()
{
  // no digitalRead here, as it is unbearable slow
  uint8_t data = 0;
  // RE low (PA9)
  GPIOA->regs->BSRR = 2 << 24;
  // RE access time (tREA), max. 30ns
  asm("NOP");
  asm("NOP");
  // fetch data from PB8 - PB15
  data = ((GPIOB->regs->IDR & 0xFF00) >> 8) & 0xff;
  // RE high (PA9)
  GPIOA->regs->BSRR = 2 << 8;
  asm("NOP");
  asm("NOP");
  return data;
}

void nandWriteDataByte(uint8_t data)
{
  digitalWrite(WE, 0);
  // apply data to PB8 - PB15
  GPIOB->regs->BSRR = data << 8;
  GPIOB->regs->BSRR = (~data) << 24;
  digitalWrite(WE, 1);
}

void nandWriteCommandByte(uint8_t cmd)
{
  digitalWrite(CLE, 1);
  nandWriteDataByte(cmd);
  digitalWrite(CLE, 0);
}

void nandWriteAddressByte(uint8_t addr)
{
  digitalWrite(ALE, 1);
  nandWriteDataByte(addr);
  digitalWrite(ALE, 0);
}

void nandWriteAddress(uint32_t unRow, uint32_t unColumn)
{
  digitalWrite(ALE, 1);
  for(uint32_t i = 0; i < g_unAddrBytes; i++)
  {
    uint8_t ab = 0;
    if(i < g_unColAddrBytes)
    {
      ab = (unColumn >> (i * 8)) & 0xff;
    }
    else
    {
      ab = (unRow >> ((i - g_unColAddrBytes) * 8)) & 0xff;      
    }
    nandWriteDataByte(ab);
  }
  digitalWrite(ALE, 0);
}

void nandWriteAddress(uint32_t unBlock, uint32_t unPage, uint32_t unColumn)
{
  nandWriteAddress((unBlock << g_unRowPageAddrBits) | unPage, unColumn);
}

void nandSetDataDirOut()
{
  pinMode(IO0, OUTPUT);
  pinMode(IO1, OUTPUT);
  pinMode(IO2, OUTPUT);
  pinMode(IO3, OUTPUT);
  pinMode(IO4, OUTPUT);
  pinMode(IO5, OUTPUT);
  pinMode(IO6, OUTPUT);
  pinMode(IO7, OUTPUT);
}

void nandSetDataDirIn()
{
  pinMode(IO0, INPUT);
  pinMode(IO1, INPUT);
  pinMode(IO2, INPUT);
  pinMode(IO3, INPUT);
  pinMode(IO4, INPUT);
  pinMode(IO5, INPUT);
  pinMode(IO6, INPUT);
  pinMode(IO7, INPUT);
}

